package es.ucm.fdi.grasia.faerie.adapter.internal.ubikSim;

import java.net.URI;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import ocp.service.ContextConsumer;
import ocp.service.ContextEntityItems;
import ocp.service.ContextItem;
import ocp.service.ContextItemBoolean;
import ocp.service.ContextItemDatetime;
import ocp.service.ContextItemFloat;
import ocp.service.ContextItemInt;
import ocp.service.ContextItemRelation;
import ocp.service.ContextItemString;
import ocp.service.ContextProducer;
import ocp.service.ContextService;

import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.ReferencePolicy;
import org.apache.felix.scr.annotations.References;
import org.apache.felix.scr.annotations.Service;
import org.apache.log4j.Logger;

import es.ucm.fdi.grasia.faerie.core.AbstractAttributeUpdaterComponent;
import es.ucm.fdi.grasia.faerie.core.Attribute;
import es.ucm.fdi.grasia.faerie.core.ContextEvent;
import es.ucm.fdi.grasia.faerie.core.ContextObserver;
import es.ucm.fdi.grasia.faerie.core.model.BasicContextValue;
import es.ucm.fdi.grasia.faerie.core.util.ContextURIs;
import es.ucm.fdi.grasia.faerie.core.util.Functions;
import es.ucm.fdi.grasia.faerie.demos.talkingAgents.model.DistanceDetected;
import es.ucm.fdi.grasia.faerie.demos.talkingAgents.model.WeightSensor;
import es.ucm.fdi.grasia.faerie.demos.talkingAgents.model.DistanceSensor;
import es.ucm.fdi.grasia.faerie.demos.talkingAgents.model.SensorState;
import es.ucm.fdi.grasia.faerie.demos.talkingAgents.model.WeightDetected;
import es.ucm.fdi.grasia.faerie.model.base.Name;

/**
 *
 * @author Chema
 */
@Service(value = ContextService.class)
@Component(inherit = true)
@References({
	@Reference(name = "contextProducer", cardinality = ReferenceCardinality.OPTIONAL_MULTIPLE,
	policy = ReferencePolicy.DYNAMIC, referenceInterface = ContextProducer.class) })
@SuppressWarnings({ "rawtypes" })
public class UbikSimAdapter extends AbstractAttributeUpdaterComponent implements ContextService {

	private Map<String, ContextProducer> contextProducers = setUpContextProducers();
	private Set<String> activeContextProducers = setUpActiveContextProducers();
	private Map<String, ContextEntityItems> contextEntityItems = setUpContextEntityItems();

	protected Map<String, ContextProducer> setUpContextProducers() {
		return new HashMap<String, ContextProducer>();
	}

	protected Set<String> setUpActiveContextProducers() {
		return new HashSet<String>();
	}

	protected Map<String, ContextEntityItems> setUpContextEntityItems() {
		return new HashMap<String, ContextEntityItems>();
	}

	@Override
	public int handle(ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			Attribute arg = event.getAttributeInstanceArg();
			if (arg != null)
				return arg.getEntity().getLocalId().toString().startsWith("/UBIK/") ? ContextObserver.HANDLE_WRITE
						: ContextObserver.HANDLE_NONE;
		}
		return ContextObserver.HANDLE_NONE;
	}

	@Override
	public synchronized void elementAdded(Attribute attributeInstance, ContextEvent event) {
		this.putUpdatedAttributeInstance(attributeInstance);

		String globalId = attributeInstance.getEntity().getName().toString();

		ContextProducer contextProducer = contextProducers.get(globalId);
		if (contextProducer != null) {
			if (activeContextProducers.add(globalId))
				Logger.getLogger(this.getClass()).debug(
						"Activating " + ContextProducer.class.getSimpleName() + ": " + globalId + " * "
						+ this.strHashCode());
			updateAttributeInstances(globalId);
		}
	}

	public synchronized void bindContextProducer(ContextProducer contextProducer, Map<String, Object> properties) {
		Logger.getLogger(this.getClass()).trace(
				"(" + contextProducer + ", " + Functions.formatProperties(properties) + ") * " + this.strHashCode());

		String globalId = (String) properties.get("id");

		contextProducers.put(globalId, contextProducer);

		if (globalId.contains("DistanceSensor")) {
			DistanceSensor sensor = (DistanceSensor) this.addReferencedEntity(new DistanceSensor(ContextURIs.
					localAuthority(), URI.create("/UBIK/"), URI.create(globalId)));

			Map<URI, ? extends Attribute> entityAttributes = this.searchAttributeInstancesByEntityId(sensor.getId());

			if (!entityAttributes.isEmpty() && activeContextProducers.add(globalId))
				Logger.getLogger(this.getClass()).debug(
						"Activating " + ContextProducer.class.getSimpleName() + ": " + globalId + " * "
						+ this.strHashCode());
		} else if (globalId.contains("FloorTileWithPressureSensor")) {
			WeightSensor sensor = (WeightSensor) this.addReferencedEntity(new WeightSensor(ContextURIs.localAuthority(),
					URI.create("/UBIK/"), URI.create(globalId)));

			Map<URI, ? extends Attribute> entityAttributes = this.searchAttributeInstancesByEntityId(sensor.getId());

			if (!entityAttributes.isEmpty() && activeContextProducers.add(globalId))
				Logger.getLogger(this.getClass()).debug(
						"Activating " + ContextProducer.class.getSimpleName() + ": " + globalId + " * "
						+ this.strHashCode());
		}
		contextProducer.activate(this);
	}

	public synchronized void unbindContextProducer(ContextProducer contextProducer, Map<String, Object> properties) {
		Logger.getLogger(this.getClass()).trace(
				"(" + contextProducer + ", " + Functions.formatProperties(properties) + ") * " + this.strHashCode());

		String globalId = (String) properties.get("id");

		contextProducer.deactivate();
		activeContextProducers.remove(globalId);

		contextProducers.remove(globalId);
	}

	@Override
	public synchronized void elementRemoved(Attribute attributeInstance, ContextEvent event) {
		Logger.getLogger(this.getClass()).trace("(" + attributeInstance + ", " + event + ") * " + this.strHashCode());

		String globalId = attributeInstance.getEntity().getName().toString();

		ContextProducer contextProducer = contextProducers.get(globalId);
		if (contextProducer != null) {
			URI entityId = ContextURIs.localEntityURI("/UBIK/", globalId);

			Map<URI, ? extends Attribute> entityAttributes = this.searchAttributeInstancesByEntityId(entityId);

			if (entityAttributes.size() == 1)
				activeContextProducers.remove(globalId);
		}

		this.removeUpdatedAttributeInstance(attributeInstance);
	}

	private void updateAttributeInstances(String globalId) {
		URI entityId = ContextURIs.localEntityURI("/UBIK/", globalId);

		Map<URI, ? extends Attribute> entityAttributes = this.searchAttributeInstancesByEntityId(entityId);

		for (Attribute attributeInstance : entityAttributes.values())
			if (attributeInstance instanceof Name) {
				Name name = (Name) attributeInstance;

				ContextEntityItems contextEntityItem = contextEntityItems.get(globalId);

				name.update(new BasicContextValue<String>("" + contextEntityItem.getContextItemValue("ubikName"), 1.0,
						this.getBundleEntity()));
			} else if (attributeInstance instanceof SensorState) {
				SensorState sensorState = (SensorState) attributeInstance;

				ContextEntityItems contextEntityItem = contextEntityItems.get(globalId);

				sensorState.update(new BasicContextValue<String>("" + contextEntityItem.getContextItemValue("state"),
						1.0, this.getBundleEntity()));
			} else if (attributeInstance instanceof WeightDetected) {
				WeightDetected weightDetected = (WeightDetected) attributeInstance;

				ContextEntityItems contextEntityItem = contextEntityItems.get(globalId);

				weightDetected.update(new BasicContextValue<Float>((Float) contextEntityItem.getContextItemValue(
						"weightDetected"), 1.0, this.getBundleEntity()));
			} else if (attributeInstance instanceof DistanceDetected) {
				DistanceDetected distanceDetected = (DistanceDetected) attributeInstance;

				ContextEntityItems contextEntityItem = contextEntityItems.get(globalId);

				distanceDetected.update(new BasicContextValue<Float>((Float) contextEntityItem.getContextItemValue(
						"distanceDetected"), 1.0, this.getBundleEntity()));
			}
	}

	private ContextEntityItems addToExisting(ContextEntityItems items) {
		String globalId = "UBIK." + items.getGlobalId();

		ContextEntityItems existing = contextEntityItems.get(globalId);
		if (existing == null)
			contextEntityItems.put(globalId, existing = items);
		else
			for (ContextItem item : items.getItems()) {
				Iterator<ContextItem> existingIt = existing.getItems().iterator();
				while (existingIt.hasNext())
					if (item.getPropertyName().equals(existingIt.next().getPropertyName())) {
						existingIt.remove();
						break;
					}
				existing.addContextItem(item);
			}

		if (activeContextProducers.contains(globalId))
			updateAttributeInstances(globalId);

		return existing;
	}

	@Override
	public boolean setContextItems(ContextEntityItems items) {
		ContextEntityItems existing = addToExisting(items);
		Logger.getLogger(this.getClass()).trace("existing = " + existing);
		return true;
	}

	/*
	 * Metodos que no se usan:
	 */
	@Override
	public boolean setContextItem(String entityClassName, String id, String propertyName, String valor) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean setContextItem(String string, String string1, String string2, int i) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItem(String string, String string1, String string2, int i) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean setContextItem(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItem(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean setContextItem(String string, String string1, String string2, float f) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItem(String string, String string1, String string2, float f) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItems(ContextEntityItems cei) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	@Deprecated
	public int getContextItemInt(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Integer getValueContextItemInt(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Set<Integer> getValuesContextItemInt(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemInt getContextItemInt(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	@Deprecated
	public float getContextItemFloat(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Float getValueContextItemFloat(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Set<Float> getValuesContextItemFloat(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemFloat getContextItemFloat(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	@Deprecated
	public boolean getContextItemBoolean(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Boolean getValueContextItemBoolean(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Set<Boolean> getValuesContextItemBoolean(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemBoolean getContextItemBoolean(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	@Deprecated
	public String getContextItemString(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public String getValueContextItemString(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Set<String> getValuesContextItemString(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemString getContextItemString(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	@Deprecated
	public Date getContextItemDate(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Date getValueContextItemDate(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Set<Date> getValuesContextItemDate(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemDatetime getContextItemDate(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemRelation getContextItemRelation(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemRelation getContextItemsRelation(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextItemRelation getContextItemRelation(String string, String string1, String string2, boolean bln) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItem(String entityClassName, String id, String propertyName, String valor) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean setContextItem(String string, String string1, String string2, Date date) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItem(String string, String string1, String string2, Date date) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean setContextItemRelation(String string, String string1, String string2, String string3, String string4) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeContextItemRelation(String string, String string1, String string2, String string3,
			String string4) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean removeAllContextItems(String string, String string1, String string2) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public ContextEntityItems getAllContextItems(String string, String string1) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public List<String> getContextEntityIds(String string) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean register(String string, String string1, ContextConsumer cc) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean register(String string, ContextConsumer cc) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public boolean unregister(String string, ContextConsumer cc) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public int historyGetContextEntityLenght(String string, String string1) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public List<ContextEntityItems> historyGetContextEntityAll(String string, String string1) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public Vector<ContextEntityItems> historyGetContextEntityLastN(String string, String string1, int i) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}

	@Override
	public void writeModel(String string) {
		Logger.getLogger(this.getClass()).trace("");
		throw new UnsupportedOperationException("Not supported yet.");
	}
}
